#include <time.h>
#include <multisync.h>
#include "ldap_plugin.h"

#define LDAPFILE "ldap"

extern gboolean multisync_debug;
 
void ldap_load_state(ldap_connection *conn)
{
  char *filename;
  FILE *f;
  char line[256];
                                                                                
  filename = g_strdup_printf("%s/%s%s", sync_get_datapath(conn->sync_pair),
			     (conn->conntype==CONNECTION_TYPE_LOCAL?"local":"remote"), LDAPFILE);
                                                                                
  if((f = fopen(filename, "r")))
    {
      while(fgets(line, 256, f))
	{
	  char prop[128], data[256];
                                                                                
	  if(sscanf(line, "%128s = %256[^\n]", prop, data) == 2)
	    {
	      if(!strcmp(prop, "servername"))
		{
		  conn->servername = g_strdup(data);
		}

	      if(!strcmp(prop, "serverport"))
		{
		  sscanf(data, "%d", (int*)&(conn->serverport));
		}

	      if(!strcmp(prop, "binddn"))
		{
		  conn->binddn = g_strdup(data);
		}

	      if(!strcmp(prop, "pwd"))
		{
		  conn->pwd = g_strdup(data);
		}

	      if(!strcmp(prop, "searchbase"))
		{
		  conn->searchbase = g_strdup(data);
		}

	      if(!strcmp(prop, "filter"))
		{
		  conn->filter = g_strdup(data);
		}
	    }
	}

      fclose(f);
    }

  g_free(filename);
}

void ldap_save_state(ldap_connection *conn)
{
  char *filename;
  FILE *f;
                                                                                
  filename = g_strdup_printf("%s/%s%s", sync_get_datapath(conn->sync_pair),
			     (conn->conntype==CONNECTION_TYPE_LOCAL?"local":"remote"), LDAPFILE);
                                                                                
  if((f = fopen(filename, "w")))
    {
      if(conn->servername)
	fprintf(f, "servername = %s\n", conn->servername);
                                                                                
      fprintf(f, "serverport = %d\n", (int)conn->serverport);

      if(conn->binddn)
	fprintf(f, "binddn = %s\n", conn->binddn);

      if(conn->pwd)
	 fprintf(f, "pwd = %s\n", conn->pwd);

      if(conn->searchbase)
	fprintf(f, "searchbase = %s\n", conn->searchbase);

      if(conn->filter)
	fprintf(f, "filter = %s\n", conn->filter);

      if(conn->syncdate)
	fprintf(f, "syncdate = %s\n", conn->syncdate);

      fclose(f);
    }
                                                                                
  g_free(filename);
}

ldap_connection* sync_connect(sync_pair *handle, connection_type type, sync_object_type object_types)
{
  ldap_connection* conn;
  int ret, ldapr;
  LDAP *ld = NULL;

  conn = g_malloc0(sizeof(ldap_connection));
  g_assert(conn);

  conn->sync_pair = handle;
  conn->conntype = type;
  conn->commondata.object_types = object_types;

  ldap_load_state(conn);

  if((ld = ldap_init(conn->servername, conn->serverport)) == NULL)
    {
      char msg[256];

      snprintf(msg, 256, "Couldn't connect to ldap server: %s on port: %d.", 
	       conn->servername, conn->serverport);

      sync_log(conn->sync_pair, msg, SYNC_LOG_ERROR);
    }
  else
    {
      char msg[256];

      snprintf(msg, 256, "Connected to ldap server: %s on port %d.", 
	       conn->servername, conn->serverport);

      sync_log(conn->sync_pair, msg, SYNC_LOG_SUCCESS);
    }

  ldap_enable_cache(ld, 3600, 0);

  conn->ld = ld;

  if(((ldapr = ldap_bind_s(ld, conn->binddn, conn->pwd, LDAP_AUTH_SIMPLE)) != LDAP_SUCCESS))
    {
      char msg[256];

      switch(ldapr)
	{
	default:
	  snprintf(msg, 256, "Couldn't bind to LDAP server, error %d.", ldapr);
	  break;
	}
      
      sync_log(conn->sync_pair, msg, SYNC_LOG_ERROR);
    }

  sync_set_requestdone(conn->sync_pair);

  return(conn);
}

void sync_disconnect(ldap_connection *conn)
{
  sync_pair *sync_pair = conn->sync_pair;
  int ldapr;

  if(conn->ld)
    {
      ldap_destroy_cache(conn->ld);
      ldapr = ldap_unbind(conn->ld);
      conn->ld = NULL;
    }

  if(ldapr != LDAP_SUCCESS)
    {
      sync_log(conn->sync_pair, "Couldn't unbind LDAP server", SYNC_LOG_ERROR);
    }
  else
    {
      sync_log(conn->sync_pair, "Disconnected from LDAP server", SYNC_LOG_SUCCESS);
    }

  sync_set_requestdone(sync_pair);
}

void sync_done(ldap_connection *conn, gboolean success)
{
  if(success)
    {
      if(conn->commondata.object_types & SYNC_OBJECT_TYPE_PHONEBOOK)
	{
	  int ret;
	  int count = -1;
	  LDAPMessage* res;
	  char *attrs[] = {"cn", "modifyTimestamp", NULL};
	  char *szBuff[512];
                                                                                
	  FILE *f;

	  char *filename = g_strdup_printf("%s/%s%s", sync_get_datapath(conn->sync_pair),
					   (conn->conntype==CONNECTION_TYPE_LOCAL?"local":"remote"), "state");
                            
	  ret = ldap_search_s(conn->ld, conn->searchbase, LDAP_SCOPE_ONELEVEL, conn->filter, attrs, 0, &res);

	  if(ret == 0)
	    {
	      count = ldap_count_entries(conn->ld, res);

	      if((f = fopen(filename, "w")))
		{
		  
		  if(count > 0)
		    {
		      LDAPMessage *entry;
		      char ** vals;
		  
		      entry = ldap_first_entry(conn->ld, res);
                                                                                
		      do
			{
			  char *modstamp = NULL;
			  GString *moddn = g_string_new("cn=");

			  if((vals = ldap_get_values(conn->ld, entry, "modifyTimestamp")) != NULL)
			    modstamp = (char*)strdup(vals[0]);

			  if((vals = ldap_get_values(conn->ld, entry, "cn")) != NULL)
			    g_string_append(moddn, vals[0]);

			  g_string_append(moddn, ",");
			  g_string_append(moddn, conn->searchbase);


			  dd(printf("Saving: %s = %s\n", modstamp, moddn->str));
			  fprintf(f, "%s = %s\n", modstamp, moddn->str);

			  free(modstamp);
			  g_string_free(moddn, FALSE);
			}
		      while((entry = ldap_next_entry(conn->ld, entry)) != NULL);
		    }

		  fclose(f);
		}
                                                                                
	      ldap_uncache_request(conn->ld, ret);
	    }
	  
	  g_free(filename);
	}
    }
  
  sync_set_requestdone(conn->sync_pair);
}

typedef struct
{
  ldap_connection *conn;
  sync_object_type newdbs;
} ldap_get_changes_arg;

GList* ldap_do_get_deleted(gpointer data)
{
  ldap_get_changes_arg *arg = data;
  ldap_connection *conn = arg->conn;
  sync_object_type newdbs = arg->newdbs;

  LDAPMessage *entry;
  FILE *f;
  char *filename = NULL;
  char ** vals;
  char *del_attrs[] = {"*", "modifyTimestamp", NULL};
  int ret;
  int count = -1;
  LDAPMessage* res;
  char *szBuff[512];
  GList *changes = NULL;
  GString *tmpbuf = g_string_new("");

  filename = g_strdup_printf("%s/%s%s", sync_get_datapath(conn->sync_pair),
			     (conn->conntype==CONNECTION_TYPE_LOCAL?"local":"remote"), "state");
                            

  //printf("ldap_do_get_deleted()\n");

  // First of all, lets find all entry that have been deleted by matching against the local database
  if((f = fopen(filename, "r")))
    {
      char line[256];
      long currpos = 0L;

      while(fgets(line, 256, f))
	{
	  char modstamp[128], moddn[256];

	  
	  
	  //printf("Read line:\n----\n%s\n", line);
	  int novals = sscanf(line, "%15s = %256[^\n]", modstamp, moddn);
	  if(novals == 2)
	    {
	      char tmp[256];
	      GString *filter;

	      sscanf(moddn, "%256[^,]%*s", tmp);

	      filter = g_string_new("(");
	      g_string_append(filter, tmp);
	      g_string_append(filter, "*)");

	      //printf("filter: %s", filter->str);

	      ret = ldap_search_s(conn->ld, conn->searchbase, LDAP_SCOPE_ONELEVEL, filter->str, del_attrs, 0, &res);
	      count = ldap_count_entries(conn->ld, res);
	      
	      if(count < 1)
		{
		  dd(printf("Couldn't find the entry: %s in the LDAP server (deleted)\n", filter->str));

		  changed_object *change = g_malloc0(sizeof(changed_object));
		  g_assert(change);
		  

		  sscanf(moddn, "cn=%256[^,]%*s", tmp);

		  int n=0;
		  for(n=0; n<strlen(tmp); n++)
		    if(tmp[n] == ' ')
		      tmp[n] = '_';

		  change->uid = (char*)strdup(tmp);
		  change->comp = (char*)strdup("");
		  change->change_type = SYNC_OBJ_HARDDELETED;
		  change->object_type = SYNC_OBJECT_TYPE_PHONEBOOK;

		  changes = g_list_append(changes, change);
		}
	      else
		{
		  dd(printf("This is still present: %s\n", filter->str));
		  g_string_append(tmpbuf, line);
		}
	      
	      ldap_uncache_request(conn->ld, ret);
	      g_string_free(filter, FALSE);
	      
	    }
	  else
	    {
	      dd(printf("Something went wrong while scanning for values! Recieved\nnovals: %d\nmodstamp: %s\nmoddn: %s\n", novals, modstamp, moddn));

	      g_string_append(tmpbuf, line);
	    }
	}

      fclose(f);
    }
  else
    {
      dd(printf("Couldn't open: %s\n", filename));
    }

  g_free(filename);

  return(changes);
}

GList* ldap_do_get_changes(gpointer data)
{
  ldap_get_changes_arg *arg = data;
  ldap_connection *conn = arg->conn;
  sync_object_type newdbs = arg->newdbs;


  GList *changes = NULL;
  change_info *chinfo;

  changed_object *change;

  int ret;
  int count = -1;
  LDAPMessage* res;
  char *attrs[] = {"*", "creatorsName", "createTimestamp", "modifiersName", "modifyTimestamp", NULL};
  char *szBuff[512];
                                                                                

  if(conn->commondata.object_types & SYNC_OBJECT_TYPE_PHONEBOOK)
    {
      FILE *f;

      char *filename = g_strdup_printf("%s/%s%s", sync_get_datapath(conn->sync_pair),
				       (conn->conntype==CONNECTION_TYPE_LOCAL?"local":"remote"), "state");
                            
      ret = ldap_search_s(conn->ld, conn->searchbase, LDAP_SCOPE_SUBTREE, conn->filter, attrs, 0, &res);

      count = ldap_count_entries(conn->ld, res);

      if(count > 0)
        {
	  LDAPMessage *entry;
	  char ** vals;
	  char *sn = NULL, *givenname = NULL, *modstamp = NULL;
	  
	  entry = ldap_first_entry(conn->ld, res);
                                                                                
	  do
	    {
	      char *lastmodutc = NULL;
	      char *uid;
	      int i=0;
	      GString *outcard;
	      GString *dn = NULL;
	      char *cn = NULL;

	      gboolean bsync = TRUE;
	      gboolean newentry = TRUE;
		  
	      
	      //printf("Found entry:\n");

	      if((vals = ldap_get_values(conn->ld, entry, "modifyTimestamp")) != NULL)
		{
		  lastmodutc = (char*)strdup(vals[0]);
		  
		  
		  if((f = fopen(filename, "r")))
		    {
		      char line[256];
		      while(fgets(line, 256, f))
			{
			  char savedmodstamp[128], savedmoddn[256];
			  
			  //printf("Read line:\n----\n%s\n", line);
			  int novals = sscanf(line, "%15s = %256[^\n]", savedmodstamp, savedmoddn);

			  if(novals == 2)
			    {
			      //printf("moddn: %s\n", moddn);
			      
			      if((vals = ldap_get_values(conn->ld, entry, "cn")) != NULL)
				{
				  GString *tmpdn = g_string_new("");
				  g_string_append(tmpdn, "cn=");
				  g_string_append(tmpdn, vals[0]);
				  g_string_append(tmpdn, ",");
				  g_string_append(tmpdn, conn->searchbase);

				  //printf("\nsavedmoddn: %s\ntmpdn: %s\nsavedmodstamp: %s\nlastmodutc: %s\n", savedmoddn, tmpdn->str, savedmodstamp, lastmodutc);

				  if(strcmp(savedmoddn, tmpdn->str) == 0)
				    {
				      newentry = FALSE;
		  

				      dd(printf("OK, the dn matches!!\n"));
				      // This is the correct dn, lets see if we need to sync
				      
				      if(strcmp(savedmodstamp, lastmodutc) == 0)
					{
					  //printf("Not syncing: %s", tmpdn->str);
					  dd(printf("Not syncing\n"));
					  bsync = FALSE;
					}
				      else
					{
					  //printf("Syncing: %s", tmpdn->str);
					  dd(printf("Syncing\n"));
					  bsync = TRUE;
					}

				      g_string_free(tmpdn, FALSE);

				      break;
				    }

				  g_string_free(tmpdn, FALSE);
				}
			    }
			}

		      fclose(f);
		    }

		  if(bsync == FALSE)
		    {

		  /*
		      free((void*)modstamp);
		      free((void*)modtime);
		  */
		      continue;
		    }


		  free((void*)lastmodutc);
		}
	      else
		{
		  dd(printf("Couldn't find modified timestamp, continuing"));
	     
		  free((void*)lastmodutc);

		  continue;
		}


	      outcard = g_string_new("");


	      g_string_append(outcard, "BEGIN:VCARD\r\n");
	      g_string_append(outcard, "VERSION:2.1\r\n");



	      if((vals = ldap_get_values(conn->ld, entry, "cn")) != NULL)
		{
		  cn = (char *)strdup(vals[0]);

		  dn = g_string_new("");
		  g_string_append(dn, "cn=");
		  g_string_append(dn, cn);
		  g_string_append(dn, ",");
		  g_string_append(dn, conn->searchbase);


		  int n=0;
		  uid = (char *)strdup(cn);

		  for(n=0; n<strlen(uid); n++)
		    {
		      if(uid[n] == ' ')
			uid[n] = '_';
		    }

		  //printf("===========\nUID: %s\n============\n", uid);
		}
	      else
		{
		  dd(printf("---------------cn returned null\n"));
		}


	      // Add FN: (name) data 
	      g_string_append(outcard, "FN:");
	      
	      // Add cn property
	      if((vals = ldap_get_values(conn->ld, entry, "cn")) != NULL)
		{
		  if(vals[0] != NULL)
		    g_string_append(outcard, vals[0]);
		}
		
	      g_string_append(outcard, "\r\n");


	      // Add N: (name) data 
	      g_string_append(outcard, "N:");
	      
	      // Add sn property
	      if((vals = ldap_get_values(conn->ld, entry, "sn")) != NULL)
		{
		  if(vals[0] != NULL)
		    g_string_append(outcard, vals[0]);
		}
		
	      g_string_append(outcard, ";");

	      // Add givenName property
	      if((vals = ldap_get_values(conn->ld, entry, "givenName")) != NULL)
		{
		  if(vals[0] != NULL)
		    g_string_append(outcard, vals[0]);
		}
	      else
		{
		  // No given name in LDAP, try to figure one out by looking at the commonName
		  if((vals = ldap_get_values(conn->ld, entry, "cn")) != NULL)
		    {
		      if(vals[0] != NULL)
			{
			  char*tmp = g_strdup(vals[0]);
			  int i=0;
			  for(i=0; i<strlen(vals[0]); i++)
			    if(tmp[i] == ' ')
			      tmp[i] = '\0';

			  g_string_append(outcard, tmp);
			  g_free(tmp);
			}
		    }
		}
		  
	      g_string_append(outcard, "\r\n");


	      // Add sn property
	      if((vals = ldap_get_values(conn->ld, entry, "fileAs")) != NULL)
		{
		  // Add FILE-AS: (name) data 
		  g_string_append(outcard, "FILE-AS:");
		  
		  if(vals[0] != NULL)
		    g_string_append(outcard, vals[0]);
		  
		  g_string_append(outcard, "\r\n");
		}
		


	      if((vals = ldap_get_values(conn->ld, entry, "mail")) != NULL)
		{
		  i=0;
		  while(vals[i] != NULL)
		    {
		      g_string_append(outcard, "EMAIL;");
		      
		      //if(i==0)
		      //g_string_append(outcard, "PREF;");

		      g_string_append(outcard, "INTERNET:");

		      g_string_append(outcard, vals[i]);
		      g_string_append(outcard, "\r\n");

		      i++;
		    }
		}
                                                                                
	      if((vals = ldap_get_values(conn->ld, entry, "mobile")) != NULL)
		{
		  i=0;
		  while(vals[i] != NULL)
		    {
		      g_string_append(outcard, "TEL;CELL:");
		      g_string_append(outcard, vals[i]);
		      g_string_append(outcard, "\r\n");

		      i++;
		    }
		}

	      if((vals = ldap_get_values(conn->ld, entry, "birthDate")) != NULL)
		{
		  if(vals[0] != NULL)
		    {
		      g_string_append(outcard, "BDAY:");
		      g_string_append(outcard, vals[0]);
		      g_string_append(outcard, "\r\n");
		      //printf("\tbirthDate=%s\n", vals[i]);
		    }
		}

	      if((vals = ldap_get_values(conn->ld, entry, "internationaliSDNNumber")) != NULL)
		{
		  if(vals[0] != NULL)
		    {
		      g_string_append(outcard, "TEL;ISDN:");
		      g_string_append(outcard, vals[0]);
		      g_string_append(outcard, "\r\n");
		      //printf("\tbirthDate=%s\n", vals[i]);
		    }
		}

	      if((vals = ldap_get_values(conn->ld, entry, "telephoneNumber")) != NULL)
		{
		  i=0;
		  while(vals[i] != NULL)
		    {
		      g_string_append(outcard, "TEL;WORK:");
		      g_string_append(outcard, vals[i]);
		      g_string_append(outcard, "\r\n");
		      //printf("\ttelephoneNumber=%s\n", vals[i]);
		      i++;
		    }
		}

	      if((vals = ldap_get_values(conn->ld, entry, "homePhone")) != NULL)
		{
		  i=0;
		  while(vals[i] != NULL)
		    {
		      g_string_append(outcard, "TEL;HOME:");
		      g_string_append(outcard, vals[i]);
		      g_string_append(outcard, "\r\n");

		      i++;
		    }
		}

	      if((vals = ldap_get_values(conn->ld, entry, "facsimileTelephoneNumber")) != NULL)
		{
		  i=0;
		  while(vals[i] != NULL)
		    {
		      g_string_append(outcard, "TEL;FAX:");
		      g_string_append(outcard, vals[i]);
		      g_string_append(outcard, "\r\n");

		      i++;
		    }
		}

	      g_string_append(outcard, "ORG:");

	      if((vals = ldap_get_values(conn->ld, entry, "o")) != NULL)
		{
		  if(vals[0] != NULL)
		      g_string_append(outcard, vals[0]);
		}

	      if((vals = ldap_get_values(conn->ld, entry, "ou")) != NULL)
		{
		  if(vals[0] != NULL)
		    {
		      g_string_append(outcard, ";");
		  
		      g_string_append(outcard, vals[0]);
		    }
		}

	      g_string_append(outcard, "\r\n");














	      if((vals = ldap_get_values(conn->ld, entry, "category")) != NULL)
		{
		  g_string_append(outcard, "CATEGORIES:");

		  i=0;
		  while(vals[i] != NULL)
		    {
		      if(i != 0)
			g_string_append(outcard, ",");

		      g_string_append(outcard, vals[i]);

		      i++;
		    }

		  g_string_append(outcard, "\r\n");
		}


	      if((vals = ldap_get_values(conn->ld, entry, "businessRole")) != NULL)
		{
		  // Add FILE-AS: (name) data 
		  g_string_append(outcard, "ROLE:");
		  
		  if(vals[0] != NULL)
		    g_string_append(outcard, vals[0]);
		  
		  g_string_append(outcard, "\r\n");
		}

	      if((vals = ldap_get_values(conn->ld, entry, "homePostalAddress")) != NULL)
		{
		  // Add FILE-AS: (name) data 
		  g_string_append(outcard, "LABEL;QUOTED-PRINTABLE;HOME:");

		  if(vals[0] != NULL)
		    {
		      GString *tmpstr = g_string_new(vals[0]);
		      //g_string_append(outcard, vals[0]);

		      int n=0;
		      for(n = 0; n<strlen(tmpstr->str); n++)
			{
			  if(tmpstr->str[n] == '\n')
			    {
			      //tmpstr->str[n] = ';';
			      g_string_erase(tmpstr, n, 1);
			      g_string_insert(tmpstr, n, "=0A");
			    }
			  else if(tmpstr->str[n] == '\r')
			    tmpstr->str[n] = '|';
			}

		      g_string_append(outcard, tmpstr->str);
		      g_string_free(tmpstr, FALSE);
		    }

		  g_string_append(outcard, "\r\n");
		}

	      if((vals = ldap_get_values(conn->ld, entry, "postalAddress")) != NULL)
		{
		  // Add FILE-AS: (name) data 
		  g_string_append(outcard, "LABEL;QUOTED-PRINTABLE;WORK:");
		  
		  if(vals[0] != NULL)
		    {
		      GString *tmpstr = g_string_new(vals[0]);
		      //g_string_append(outcard, vals[0]);

		      int n=0;
		      for(n = 0; n<strlen(tmpstr->str); n++)
			{
			  if(tmpstr->str[n] == '\n')
			    {
			      //tmpstr->str[n] = ';';
			      g_string_erase(tmpstr, n, 1);
			      g_string_insert(tmpstr, n, "=0A");
			    }
			  else if(tmpstr->str[n] == '\r')
			    tmpstr->str[n] = '|';
			}

		      g_string_append(outcard, tmpstr->str);
		      g_string_free(tmpstr, FALSE);
		    }
		  
		  g_string_append(outcard, "\r\n");
		}

	      if((vals = ldap_get_values(conn->ld, entry, "spouseName")) != NULL)
		{
		  // Add FILE-AS: (name) data 
		  g_string_append(outcard, "X-EVOLUTION-SPOUSE:");
		  
		  if(vals[0] != NULL)
		    g_string_append(outcard, vals[0]);
		  
		  g_string_append(outcard, "\r\n");
		}

	      if((vals = ldap_get_values(conn->ld, entry, "assistantName")) != NULL)
		{
		  // Add FILE-AS: (name) data 
		  g_string_append(outcard, "X-EVOLUTION-ASSISTANT:");
		  
		  if(vals[0] != NULL)
		    g_string_append(outcard, vals[0]);
		  
		  g_string_append(outcard, "\r\n");
		}

	      if((vals = ldap_get_values(conn->ld, entry, "title")) != NULL)
		{
		  // Add FILE-AS: (name) data 
		  g_string_append(outcard, "TITLE:");
		  
		  if(vals[0] != NULL)
		    g_string_append(outcard, vals[0]);
		  
		  g_string_append(outcard, "\r\n");
		}

	      if((vals = ldap_get_values(conn->ld, entry, "labeledURI")) != NULL)
		{
		  // Add FILE-AS: (name) data 
		  g_string_append(outcard, "URL:");
		  
		  if(vals[0] != NULL)
		    g_string_append(outcard, vals[0]);
		  
		  g_string_append(outcard, "\r\n");
		}

	      if((vals = ldap_get_values(conn->ld, entry, "managerName")) != NULL)
		{
		  // Add FILE-AS: (name) data 
		  g_string_append(outcard, "X-EVOLUTION-MANAGER:");
		  
		  if(vals[0] != NULL)
		    g_string_append(outcard, vals[0]);
		  
		  g_string_append(outcard, "\r\n");
		}

	      if((vals = ldap_get_values(conn->ld, entry, "anniversary")) != NULL)
		{
		  // Add FILE-AS: (name) data 
		  g_string_append(outcard, "X-EVOLUTION-ANNIVERSARY:");
		  
		  if(vals[0] != NULL)
		    g_string_append(outcard, vals[0]);
		  
		  g_string_append(outcard, "\r\n");
		}

	      if((vals = ldap_get_values(conn->ld, entry, "note")) != NULL)
		{
		  // Add FILE-AS: (name) data 
		  g_string_append(outcard, "NOTE;QUOTED-PRINTABLE:");
		  
		  if(vals[0] != NULL)
		    {
		      GString *tmpstr = g_string_new(vals[0]);
		      //g_string_append(outcard, vals[0]);

		      int n=0;
		      for(n = 0; n<strlen(tmpstr->str); n++)
			{
			  if(tmpstr->str[n] == '\n')
			    {
			      g_string_erase(tmpstr, n, 1);
			      g_string_insert(tmpstr, n, "=0A");
			    }
			}

		      g_string_append(outcard, tmpstr->str);
		      g_string_free(tmpstr, FALSE);
		    }

		  g_string_append(outcard, "\r\n");
		}

	      if((vals = ldap_get_values(conn->ld, entry, "roomNumber")) != NULL)
		{
		  // Add FILE-AS: (name) data 
		  g_string_append(outcard, "X-EVOLUTION-OFFICE:");
		  
		  if(vals[0] != NULL)
		    g_string_append(outcard, vals[0]);
		  
		  g_string_append(outcard, "\r\n");
		}

	      if((vals = ldap_get_values(conn->ld, entry, "displayName")) != NULL)
		{
		  // Add FILE-AS: (name) data 
		  g_string_append(outcard, "NICKNAME:");
		  
		  if(vals[0] != NULL)
		    g_string_append(outcard, vals[0]);
		  
		  g_string_append(outcard, "\r\n");
		}


	      //g_string_append(outcard, "REV:");
	      //g_string_append(outcard, (char*)sync_timet_to_dt(modtime));
	      //g_string_append(outcard, "\r\n");

	      g_string_append(outcard, "END:VCARD\r\n");

	      change = g_malloc0(sizeof(changed_object));
	      g_assert(change);
	      change->uid = g_strdup(uid);
	      free(uid);
	      change->comp = g_strdup(outcard->str);
	      change->change_type = ((newentry == TRUE) ? SYNC_OBJ_ADDED : SYNC_OBJ_MODIFIED);
	      change->object_type = SYNC_OBJECT_TYPE_PHONEBOOK;

	      g_string_free(outcard, TRUE);
	      changes = g_list_append(changes, change);



	      //	      if(modtime != NULL)
	      //		free((void*)modtime);


	      g_string_free(dn, FALSE);

	      if(modstamp != NULL)
		free((void*)modstamp);

	      if(lastmodutc != NULL)
		free((void*)lastmodutc);

	      if(cn != NULL)
		free(cn);
	    }
	  while((entry = ldap_next_entry(conn->ld, entry)) != NULL);
        }
                                                                                
      ldap_uncache_request(conn->ld, ret);
      g_free(filename);
    }

  return(changes);
}

void get_changes(ldap_connection *conn, sync_object_type newdbs)
{
  ldap_get_changes_arg *arg;

  arg = g_malloc0(sizeof(ldap_get_changes_arg));
  arg->conn = conn;
  arg->newdbs = newdbs;

  GList *changes = NULL;

  

  changes = ldap_do_get_deleted(arg);
  changes = g_list_concat(changes, ldap_do_get_changes(arg));

  g_free(arg);

  change_info *chinfo;

  chinfo = g_malloc(sizeof(change_info));
  chinfo->changes = changes;
  chinfo->newdbs = 0;
  sync_set_requestdata(chinfo, conn->sync_pair);
}

void save_ldap_entry(ldap_connection *conn,
                             char *comp,
                             char *uid, sync_object_type objtype,
                             char *uidret, int *uidretlen) 
{
  LDAPMod **attrs, **tmpattrs;
  GString * dn = NULL;
                                                                                
  char *data;
  char *cn = NULL;
  char *sn = NULL;
  char **values;
  int currattr = 0;
  int retval;

  if (!uid && !comp) 
    {
      sync_set_requestfailed(conn->sync_pair);
      return;
    }


  dd(printf("save_ldap_entry()\nuid: %s\nuidret: %s\n", uid, uidret));

  values = malloc(6 * sizeof(char*));

  values[0] = (char*)malloc(50);
  snprintf(values[0], 50, "top");

  values[1] = (char*)malloc(50);
  snprintf(values[1], 50, "person");

  values[2] = (char*)malloc(50);
  snprintf(values[2], 50, "organizationalPerson");

  values[3] = (char*)malloc(50);
  snprintf(values[3], 50, "inetOrgPerson");

  values[4] = (char*)malloc(50);
  snprintf(values[4], 50, "evolutionPerson");

  values[5] = NULL;

  attrs = malloc((currattr+1)*sizeof(LDAPMod *));
  attrs[0] = (LDAPMod*)malloc(sizeof(LDAPMod));
  attrs[0]->mod_op = LDAP_MOD_ADD;
  attrs[0]->mod_type = "objectclass";
  attrs[0]->mod_values = values;
  currattr++;


  
  data = sync_get_key_data(comp, "FN");
  if(data != NULL && strlen(data)>0)
    {
      //printf("FN returned: %s", data);
      cn = (char*)g_strdup(data);
      g_free(data);
    }

  data = sync_get_key_data(comp, "N");
  if(data != NULL)
    {
      dd(printf("Got N=%s\n", data));
      int x=0, firstname=0, lastname=0;
      int maxlen = strlen(data);
      for(x=0; x<maxlen; x++)
	{
	  if(data[x] == ';')
	    {
	      data[x] = '\0';
	      
	      if(x < maxlen && firstname == 0)
		{
		  firstname = x+1;
		  break;
		}
	    }
	  else if(data[x] == '\0' && firstname == 0)
	    {
	      firstname = x+1;
	      break;
	    }
	}

      GString *tmp = g_string_new("");

      g_string_append(tmp, (char*)&data[firstname]);
      g_string_append(tmp, " ");
      g_string_append(tmp, (char*)&data[lastname]);
      sn = (char*)g_strdup((char*)&data[lastname]);

      if(cn == NULL)
	cn = (char*)g_strdup(tmp->str);

      g_string_free(tmp, FALSE);
      g_free(data);
    }


  /*
  printf("sn: %s\ncn: %s\n", sn, cn);
  printf("finished forming names\n");
  */
  dn = g_string_new("cn=");
  g_string_append(dn, cn);
  g_string_append(dn, ",");
  g_string_append(dn, conn->searchbase);

  /*
  printf("save_ldap_entry()\n");
  printf("dn:  %s\n", dn->str);
  printf("\nuid: %s\nluid: %s\ncomp: %s\n", uid, uidret, comp);
  */


  // Parsa org
  data = sync_get_key_data(comp, "ORG");
  if(data != NULL)
    {
      char o[256], ou[256];

      dd(printf("Got ORG: %s\n", data));

      sscanf(data, "%[^;];%[^\n]", o, ou);

      dd(printf("Broke down to\no: %s\nou: %s\n", o, ou));

      if(o != NULL)
	{
	  values = g_malloc0(2 * sizeof(char*));
	  values[0] = (char*)o;
	  values[1] = NULL;

	  attrs = add_ldap_attribute(attrs, currattr++, "o", values);
	}

      if( ou != NULL)
	{
	  values = g_malloc0(2 * sizeof(char*));
	  values[0] = (char*)ou;
	  values[1] = NULL;

	  attrs = add_ldap_attribute(attrs, currattr++, "ou", values);
	}

      g_free(data);
    }



  //data = sync_get_key_data(comp, "EMAIL");
  values =  sync_get_all_key_data(comp, "EMAIL");

  if(values != NULL)
    {
      attrs = add_ldap_attribute(attrs, currattr++, "mail", values);
    }

  data = sync_get_key_data(comp, "TEL;CELL");

  if(data != NULL)
    {
      values = g_malloc0(2 * sizeof(char*));
      values[0] = (char*)strdup(data);
      values[1] = NULL;

      g_free(data);

      attrs = add_ldap_attribute(attrs, currattr++, "mobile", values);
    }
  
  data = sync_get_key_data(comp, "TEL;ISDN");

  if(data != NULL)
    {
      values = g_malloc0(2 * sizeof(char*));
      values[0] = (char*)strdup(data);
      values[1] = NULL;
      
      g_free(data);
      
      attrs = add_ldap_attribute(attrs, currattr++, "internationaliSDNNumber", values);
    }

  data = sync_get_key_data(comp, "TEL;WORK");

  if(data != NULL)
    {
      values = g_malloc0(2 * sizeof(char*));
      values[0] = (char*)strdup(data);
      values[1] = NULL;

      g_free(data);

      attrs = add_ldap_attribute(attrs, currattr++, "telephoneNumber", values);
    }

  data = sync_get_key_data(comp, "TEL;FAX");

  if(data != NULL)
    {
      values = g_malloc0(2 * sizeof(char*));
      values[0] = (char*)strdup(data);
      values[1] = NULL;

      g_free(data);

      attrs = add_ldap_attribute(attrs, currattr++, "facsimileTelephoneNumber", values);
    }


  data = sync_get_key_data(comp, "TEL;HOME");

  if(data != NULL)
    {
      values = g_malloc0(2 * sizeof(char*));
      values[0] = (char*)strdup(data);
      values[1] = NULL;

      g_free(data);

      attrs = add_ldap_attribute(attrs, currattr++, "homePhone", values);
    }



  data = sync_get_key_data(comp, "CATEGORIES");

  if(data != NULL)
    {
      values = g_malloc0(2 * sizeof(char*));
      values[0] = (char*)strdup(data);
      values[1] = NULL;

      g_free(data);

      attrs = add_ldap_attribute(attrs, currattr++, "category", values);
    }


  data = sync_get_key_data(comp, "ROLE");

  if(data != NULL)
    {
      values = g_malloc0(2 * sizeof(char*));
      values[0] = (char*)strdup(data);
      values[1] = NULL;

      g_free(data);

      attrs = add_ldap_attribute(attrs, currattr++, "businesRole", values);
    }

  //data = sync_get_key_data(comp, "ADR;HOME");
  data = sync_get_key_data(comp, "LABEL;QUOTED-PRINTABLE;HOME");

  if(data != NULL)
    {
      /*
      char *box = NULL, *addr2 = NULL, *addr1 = NULL, *city = NULL;
      char *state=NULL, *zip=NULL, *country = NULL;

      sscanf(data, "[^;][^;][^;][^;][^;][^;]%a[^\n]", box, addr2, addr1, city, state, zip, country);
      */
      GString *tmpstr = g_string_new(data);

      int n=0;
      for(n = 0; n<strlen(tmpstr->str)-3; n++)
	{
	  if(strncmp((char*)&tmpstr->str[n], "=0A", 3) == 0)
	    {
	      g_string_erase(tmpstr, n, 3);
	      g_string_insert(tmpstr, n, "\n");
	    }
	}
      
      values = g_malloc0(2 * sizeof(char*));
      values[0] = (char*)strdup(tmpstr->str);
      values[1] = NULL;

      g_free(data);
      g_string_free(tmpstr, FALSE);

      attrs = add_ldap_attribute(attrs, currattr++, "homePostalAddress", values);
    }

  //data = sync_get_key_data(comp, "ADR;WORK");
  data = sync_get_key_data(comp, "LABEL;QUOTED-PRINTABLE;WORK");

  if(data != NULL)
    {
      GString *tmpstr = g_string_new(data);

      int n=0;
      for(n = 0; n<strlen(tmpstr->str)-3; n++)
	{
	  if(strncmp((char*)&tmpstr->str[n], "=0A", 3) == 0)
	    {
	      g_string_erase(tmpstr, n, 3);
	      g_string_insert(tmpstr, n, "\n");
	    }
	}

      values = g_malloc0(2 * sizeof(char*));
      values[0] = (char*)strdup(tmpstr->str);
      values[1] = NULL;

      g_free(data);
      g_string_free(tmpstr, FALSE);

      attrs = add_ldap_attribute(attrs, currattr++, "postalAddress", values);
    }

  data = sync_get_key_data(comp, "X-EVOLUTION-SPOUSE");

  if(data != NULL)
    {
      values = g_malloc0(2 * sizeof(char*));
      values[0] = (char*)strdup(data);
      values[1] = NULL;

      g_free(data);

      attrs = add_ldap_attribute(attrs, currattr++, "spouseName", values);
    }

  data = sync_get_key_data(comp, "X-EVOLUTION-ASSISTANT");

  if(data != NULL)
    {
      values = g_malloc0(2 * sizeof(char*));
      values[0] = (char*)strdup(data);
      values[1] = NULL;

      g_free(data);

      attrs = add_ldap_attribute(attrs, currattr++, "assistantName", values);
    }

  data = sync_get_key_data(comp, "TITLE");

  if(data != NULL)
    {
      values = g_malloc0(2 * sizeof(char*));
      values[0] = (char*)strdup(data);
      values[1] = NULL;

      g_free(data);

      attrs = add_ldap_attribute(attrs, currattr++, "title", values);
    }
  
  data = sync_get_key_data(comp, "URL");

  if(data != NULL)
    {
      values = g_malloc0(2 * sizeof(char*));
      values[0] = (char*)strdup(data);
      values[1] = NULL;

      g_free(data);

      attrs = add_ldap_attribute(attrs, currattr++, "labeledURI", values);
    }
  
  data = sync_get_key_data(comp, "X-EVOLUTION-MANAGER");

  if(data != NULL)
    {
      values = g_malloc0(2 * sizeof(char*));
      values[0] = (char*)strdup(data);
      values[1] = NULL;

      g_free(data);

      attrs = add_ldap_attribute(attrs, currattr++, "managerName", values);
    }
  
  data = sync_get_key_data(comp, "X-EVOLUTION-ANNIVERSARY");

  if(data != NULL)
    {
      values = g_malloc0(2 * sizeof(char*));
      values[0] = (char*)strdup(data);
      values[1] = NULL;

      g_free(data);

      attrs = add_ldap_attribute(attrs, currattr++, "anniversary", values);
    }

  data = sync_get_key_data(comp, "NOTE");

  if(data != NULL)
    {
      GString *tmpstr = g_string_new(data);

      int n=0;
      for(n = 0; n<strlen(tmpstr->str)-3; n++)
	{
	  if(strncmp((char*)&tmpstr->str[n], "=0A", 3) == 0)
	    {
	      g_string_erase(tmpstr, n, 3);
	      g_string_insert(tmpstr, n, "\n");
	    }
	}
      
      values = g_malloc0(2 * sizeof(char*));
      values[0] = (char*)strdup(tmpstr->str);
      values[1] = NULL;

      g_free(data);
      g_string_free(tmpstr, FALSE);

      attrs = add_ldap_attribute(attrs, currattr++, "note", values);
    }
  
  data = sync_get_key_data(comp, "X-EVOLUTION-OFFICE");

  if(data != NULL)
    {
      values = g_malloc0(2 * sizeof(char*));
      values[0] = (char*)strdup(data);
      values[1] = NULL;

      g_free(data);

      attrs = add_ldap_attribute(attrs, currattr++, "roomNumber", values);
    }

  data = sync_get_key_data(comp, "NICKNAME");

  if(data != NULL)
    {
      values = g_malloc0(2 * sizeof(char*));
      values[0] = (char*)strdup(data);
      values[1] = NULL;

      g_free(data);

      attrs = add_ldap_attribute(attrs, currattr++, "displayName", values);
    }


  values = g_malloc0(2 * sizeof(char*));
  values[0] = (char*)strdup(sn);
  values[1] = NULL;


  attrs = add_ldap_attribute(attrs, currattr++, "sn", values);


  values = g_malloc0(2 * sizeof(char*));
  values[0] = (char*)strdup(cn);
  values[1] = NULL;

  attrs = add_ldap_attribute(attrs, currattr++, "cn", values);

  tmpattrs = attrs;
  attrs = malloc((currattr+1)*sizeof(LDAPMod *));
  memcpy(attrs, tmpattrs, currattr*sizeof(LDAPMod *));
  free(tmpattrs);


  attrs[currattr] = NULL;
          
  dd(printf("About to add\n"));
  retval = ldap_add_s(conn->ld, dn->str, attrs);

  char msg[256];
  msg[0] = '\0';

  switch(retval)
    {
    case LDAP_SUCCESS:
      snprintf(msg, 256, "Added/Updated object %s\n", dn->str);
      break;

    case LDAP_OBJECT_CLASS_VIOLATION:
      snprintf(msg, 256, "Object class violation while adding %s\n", dn->str);
      break;

    case LDAP_INVALID_SYNTAX:
      snprintf(msg, 256, "Invalid LDAP syntax while adding %s\n", dn->str);
      break;

    case LDAP_ALREADY_EXISTS:
      snprintf(msg, 256, "Object already exists: %s\n", dn->str);
      break;

    default:
      snprintf(msg, 256, "LDAP operation returned with value: %d\n", retval);
      break;
    }

  if(retval == LDAP_SUCCESS)
    sync_log(conn->sync_pair, msg, SYNC_LOG_SUCCESS);
  else
    sync_log(conn->sync_pair, msg, SYNC_LOG_ERROR);
    

  int i = 0;
  while(attrs[i] != NULL)
    {
      int j = 0;
      while(attrs[i]->mod_values[j] != NULL)
	{
	  free(attrs[i]->mod_values[j]);

	  j++;
	}

      free(attrs[i]);
      i++;
    }


  memcpy(uidret, cn, strlen(cn));
  uidret[strlen(cn)] = '\0';
  *uidretlen = strlen(uidret);

  int n=0;
  for(n=0; n<strlen(uidret); n++)
    {
      if(uidret[n] == ' ')
	uidret[n] = '_';
    }

  g_string_free(dn, FALSE);

  if(cn != NULL)
    g_free(cn);

  if(sn != NULL)
    g_free(sn);

  sync_set_requestdone(conn->sync_pair);
}

LDAPMod **add_ldap_attribute(LDAPMod **attrs, int attrno, char* mod_type, char **values)
{
  LDAPMod **tmpattrs;

  tmpattrs = attrs;
  attrs = malloc((attrno+1)*sizeof(LDAPMod *));
  memcpy(attrs, tmpattrs, attrno*sizeof(LDAPMod *));
  free(tmpattrs);


  attrs[attrno] = (LDAPMod*)malloc(sizeof(LDAPMod));
  attrs[attrno]->mod_op = LDAP_MOD_ADD;
  attrs[attrno]->mod_type = mod_type;
  attrs[attrno]->mod_values = values;

  return attrs;
}

void delete_ldap_entry(ldap_connection *conn,
                             char *comp,
                             char *uid, sync_object_type objtype,
                             char *uidret, int *uidretlen) 
{
  char *cn = NULL;
  GString *entrydn = NULL;
  
  cn = (char*)strdup(uid);
  
  int n=0;
  for(n = 0; n<strlen(cn); n++)
    {
      if(cn[n] == '_')
	cn[n] = ' ';
    }

  entrydn = g_string_new("cn=");
  g_string_append(entrydn, cn);
  g_string_append(entrydn, ",");
  g_string_append(entrydn, conn->searchbase);



  //printf("\n****************delete_ldap_entry()\nuid: %s\nentrydn: %s\n***********\n", uid, entrydn->str);

  if(ldap_delete_s(conn->ld, entrydn->str) != 0)
    {
      char msg[256];

      sprintf(msg, "Couldn't delete entry: %s.", entrydn->str);

      sync_log(conn->sync_pair, msg, SYNC_LOG_ERROR);
    }


  if(cn!=NULL)
    free(cn);

  g_string_free(entrydn, FALSE);

  sync_set_requestdone(conn->sync_pair);
}

void syncobj_modify(ldap_connection *conn,
                    char *comp,
                    char *uid, sync_object_type objtype,
                    char *uidret, int *uidretlen) 
{
  //printf("syncobj_modify()\nuid: %s\nuidret: %s\n\n", uid, uidret);

  if(uid != NULL)
    delete_ldap_entry(conn, NULL, uid, objtype, NULL, NULL);

  save_ldap_entry(conn, comp, uid, objtype, uidret, uidretlen);
}
 

/*
 * Delete the object from the LDAP server. We won't care about the softdelete
 * flag since the LDAP server doesn't have soft or harddelete.
 *
 */
void syncobj_delete(ldap_connection *conn, char *uid,
                    sync_object_type objtype, int softdelete) 
{
  
  dd(printf("syncobj_delete()\nuid: %s\n", uid));
  delete_ldap_entry(conn, NULL, uid, objtype, NULL, NULL);
}
                                                                                

char* short_name()
{
  return("ldap-sync");
}

char *long_name()
{
  return("LDAP");
}

// Return the types of objects that this client handle
sync_object_type object_types() 
{
  return(SYNC_OBJECT_TYPE_PHONEBOOK);
}

void plugin_init(void) 
{
} 

char *plugin_info()
{
  return("Allows synchronization of a addressbook against a specific DN in a LDAP server.");
}

gboolean always_connected()
{
  return(FALSE);
}

int plugin_API_version(void)
{
  return(3);
}

/*
 * This method returns all values for a given key in the vCard as a array
 */
char** sync_get_all_key_data(char *card, char *key) 
{
  char *pos = card;
  int l = strlen(key);
  char *res = g_strdup("");

  char **retval;
  char **tmpretval;
  int size=0;
  retval = malloc(sizeof(char*));

  while (pos) {
    if (!strncmp(pos, key, l) && (pos[l] == ':' || pos[l] == ';')) {
      char *start, *end;
      start = strstr(pos+l, ":");
      if (start) {
	char *tmp, *line, *newstart;
	do {
	  start++;
	  end = strstr(start, "\n");
	  if (!end)
	    end = card+strlen(card);
	  newstart = end+1;
	  if (*(end-1) == '\r')
	    end--;
	  line = g_strndup(start, end-start);
	  tmp = g_strdup_printf("%s%s", res, line);
	  g_free(res);
	  g_free(line);
	  res = tmp;
	  start = newstart;
	} while(start < card+strlen(card) && start[0] == ' ');

	tmpretval = (char**)malloc((size+2)*sizeof(char*));
	int n=0;
	for(n=0; n<size;n++)
	  tmpretval[n] = retval[n];
	tmpretval[size] = res;
	res = g_strdup("");
	retval = tmpretval;
	size++;
	//return(res);
      }
    }
    pos = strstr(pos, "\n");
    if (pos)
      pos += 1;
  }
  g_free(res);

  retval[size] = NULL;

  return(retval);
}
